# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod
from importlib.metadata import entry_points

from lxml import etree

from karabo.native import (
    AccessLevel, AccessMode, Bool, Configurable, Node, String)

DEFAULT_PROTOCOL = "exist_db"


class ProjectDBError(Exception):
    pass


class DbConnectionNodeBase(Configurable):
    @abstractmethod
    def get_db(self, test_mode=False, init_db=False):
        """creates an object subclass of ``DatabaseBase``

        To Be implemented in the subclass

        :param test_mode: Defines if the DB Connection is used during a test.
                          Behaviour is implementation-specific.
        :param init_db: True if database should be initialized
        :return: an object subclass of ``DatabaseBase``
        """
        raise NotImplementedError


def get_node():
    entrypoints = list(entry_points(group="karabo.project_db"))
    protocols = [ep.name for ep in entrypoints]
    assert DEFAULT_PROTOCOL in protocols, ",".join(protocols)

    class ProjectNode(Configurable):
        _plugins = {}

        protocol = String(
            defaultValue=DEFAULT_PROTOCOL,
            options=protocols,
            accessMode=AccessMode.INITONLY)

        testMode = Bool(
            displayedName="Test Mode",
            defaultValue=False,
            requiredAccessLevel=AccessLevel.ADMIN)

        for ep in entrypoints:
            node_key = ep.name
            node_cls = ep.load()
            assert issubclass(node_cls, DbConnectionNodeBase)
            _plugins[node_key] = node_cls
            locals()[node_key] = Node(node_cls)

    return Node(ProjectNode)


def get_project(node, init_db=False):
    """creates an object subclass of ``DatabaseBase``

    :param node: the instance of ``Node`` object implementing the
                 ``ProjectNode`` class.
    :param init_db: True if database should be initialized.
    :return: an object subclass of ``DatabaseBase``
    """
    protocol = node.protocol.value
    ep_node = getattr(node, protocol)
    return ep_node.get_db(test_mode=node.testMode.value, init_db=init_db)


def to_string(xml_rep):
    """returns a string serialization of an xml element

    the output is listed on multiple lines and unicode encoded"""
    return etree.tostring(
        xml_rep,
        pretty_print=True,
        encoding="unicode",
        xml_declaration=False)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import contextlib
import os
import tempfile
from unittest import TestCase

from karabo.project_db.tests.base import ProjectDatabaseVerification

from ..database import ProjectDatabase


class TestFileProjectDatabase(ProjectDatabaseVerification, TestCase):
    @classmethod
    def setUpClass(cls):
        base_dir = os.path.join(os.environ['KARABO'], 'var', 'data')
        with contextlib.ExitStack() as stack:
            cls.test_dir = tempfile.TemporaryDirectory(dir=base_dir)
            stack.enter_context(cls.test_dir)

    def _db_init(self):
        return ProjectDatabase(
            str(self.test_dir), test_mode=True, init_db=True)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
import tempfile
from contextlib import contextmanager

from karabo.middlelayer.tests.eventloop import DeviceTest
from karabo.middlelayer_devices.project_manager import ProjectManager
from karabo.native import Hash
from karabo.project_db.file_db.database import ProjectDatabase

from .projectdb_util import ConsumerDevice, VerificationProjectManager


class TestFileProjectManager(VerificationProjectManager, DeviceTest):
    def _db_init(self):
        return ProjectDatabase(self.dir, test_mode=True, init_db=True)

    @classmethod
    @contextmanager
    def lifetimeManager(cls):
        base_dir = os.path.join(os.environ['KARABO'], 'var', 'data')
        with tempfile.TemporaryDirectory(dir=base_dir) as cls.dir:
            path = str(os.path.relpath(cls.dir, base_dir))
            conf = Hash("_deviceId_", "projManTest")
            conf["projectDB"] = Hash("protocol", "file_db",
                                     "file_db", Hash("dir", path),
                                     "testMode", True)
            cls.local = ProjectManager(conf)
            cls.consume = ConsumerDevice({"_deviceId_": "consumeTest"})
            with cls.deviceManager(cls.local, cls.consume, lead=cls.local):
                yield



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa: F401
"""
Unit-aware replacements for :py:mod:`numpy` functions
=====================================================

Most :py:mod:`numpy` functions (namely: ufuncs) support Karabo values
out-of-the-box.
Some, however, do not. Karabo defines its own drop-in replacements for those
functions in the :py:mod:`karabo.numeric` module. These are:

.. function:: karabo.numeric.linspace

    See :py:func:`numpy.linspace`

.. function:: karabo.numeric.dot

    See :py:func:`numpy.dot`

.. function:: karabo.numeric.cross

    See :py:func:`numpy.cross`
"""
from .unitutil import cross, dot, linspace



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import copy
from asyncio import iscoroutinefunction
from collections.abc import Iterable
from functools import wraps

import numpy as np

from karabo.common.states import StateSignifier as SignifierBase
from karabo.native import (
    KaraboValue, QuantityValue, newest_timestamp, wrap, wrap_function)


@wraps(np.linspace)
@wrap_function
def linspace(start, stop, *args, **kwargs):
    if not (isinstance(start, QuantityValue) or
            isinstance(stop, QuantityValue)):
        return np.linspace(start, stop, *args, **kwargs)

    if not isinstance(start, QuantityValue):
        start = QuantityValue(start)

    stop = float(stop / start.units)
    return QuantityValue(
        np.linspace(start.magnitude, stop, *args, **kwargs), start.units)


@wraps(np.dot)
@wrap_function
def dot(a, b, *args, **kwargs):
    if not (isinstance(a, QuantityValue) or isinstance(b, QuantityValue)):
        return np.dot(a, b, *args, **kwargs)

    if not isinstance(a, QuantityValue):
        a = QuantityValue(a)

    if not isinstance(b, QuantityValue):
        b = QuantityValue(b)

    return QuantityValue(np.dot(a.magnitude, b.magnitude), a.units * b.units)


@wraps(np.cross)
@wrap_function
def cross(a, b, *args, **kwargs):
    if not (isinstance(a, QuantityValue) or isinstance(b, QuantityValue)):
        return np.cross(a, b, *args, **kwargs)

    if not isinstance(a, QuantityValue):
        a = QuantityValue(a)

    if not isinstance(b, QuantityValue):
        b = QuantityValue(b)

    return QuantityValue(np.cross(a.magnitude, b.magnitude), a.units * b.units)


def maximum(iterable):
    """Return the maximum value of the iterable

    This function takes into account KaraboValues for newest timestamp
    generation.
    """
    assert isinstance(iterable, Iterable)

    ret = max(iterable)
    ret = wrap(copy.copy(ret))
    ret.timestamp = newest_timestamp(iterable)

    return ret


def minimum(iterable):
    """Return the minimum value of the iterable

    This function takes into account KaraboValues for newest timestamp
    generation.
    """
    assert isinstance(iterable, Iterable)

    ret = min(iterable)
    ret = wrap(copy.copy(ret))
    ret.timestamp = newest_timestamp(iterable)

    return ret


class StateSignifier(SignifierBase):
    """Wrapper of the StateSignifier to provide newest timestamp
    """

    def returnMostSignificant(self, iterable):
        ret = super().returnMostSignificant(iterable)
        ret = wrap(copy.copy(ret))
        ret.timestamp = newest_timestamp(iterable)

        return ret


def removeQuantity(func):
    """Decorate a function to remove QuantityValue/KaraboValue input

    This function works as well with async declarations and can be used as::

    @removeQuantity
    def calculate(x, y):
        assert not isinstance(x, KaraboValue)
        assert not isinstance(y, KaraboValue)
        return x, y

    Note: This decorator does not cast to base units! In case of booleans,
          they can be compared by identity.
    """

    def _convert_input(args, kwargs):
        n_args = [
            value.value if isinstance(value, KaraboValue) else value
            for value in tuple(args)]
        n_kwargs = {
            key: value.value if isinstance(value, KaraboValue) else value
            for key, value in kwargs.items()}
        return n_args, n_kwargs

    if iscoroutinefunction(func):

        @wraps(func)
        async def wrapper(*args, **kwargs):
            n_args, n_kwargs = _convert_input(args, kwargs)
            return await func(*n_args, **n_kwargs)
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            n_args, n_kwargs = _convert_input(args, kwargs)
            return func(*n_args, **n_kwargs)

    return wrapper



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from uuid import uuid4


def _gen_uuid():
    return str(uuid4())


def create_device(db):
    sub_uuid = _gen_uuid()
    conf_uuid1 = _gen_uuid()
    conf_uuid2 = _gen_uuid()
    xml = ('<xml item_type="{atype}" uuid="{uuid}">'
           '<device_instance active_rev="0" instance_id="{instance_id}">'
           '<device_config revision="0" uuid="{conf_uuid1}" />'
           '<device_config revision="1" uuid="{conf_uuid2}" />'
           '</device_instance>'
           '</xml>').format(uuid=sub_uuid,
                            atype='device_instance',
                            instance_id=sub_uuid,
                            conf_uuid1=conf_uuid1,
                            conf_uuid2=conf_uuid2
                            )

    db.save_item("LOCAL", sub_uuid, xml)
    return sub_uuid, conf_uuid1


def create_hierarchy(db):
    """
    Create a minimal project hierarchy representative of Karabo entries in
    ExistDB

    :param db: a `ProjectDatabase` instance
    :return: a tuple consisting of:
      - the uuid of the project created
      - a dictionary mapping device ids (key) to their active
         configuration's uuid (value)

    The following entities are created in hierarchical order, with attributes
    mentioned in parentheses. The simple_name is frequently set to the uuid,
    and uuids are randomly generated at each call::

        + Project(item_type:='project', uuid, simple_name:='Project')
        |
        +- 4x Scene(item_type:='scene', uuid, simple_name:=uuid)
        |
        +- 4x Server(item_type:='device_server', uuid, simple_name:=uuid)
           |
           +- 4x Device(item_type:='device_instance', uuid)
              |
              2x Device Config(revision=0|1, uuid)

    The Project XML has the format::

        <xml item_type="Project", uuid="...", simple_name="Project">
            <root>
                <project>
                    <scenes>
                        ...
                    </scenes>
                    <servers>
                        ...
                    </servers>
                </project>
            </root>
        </xml>

    Each scene in the project XML has a minimal XML of the form::

        <xml item_type="scene", uuid="...", simple_name="..." />

    For each scene a top-level entry is created::

        <xml item_type="scene" uuid="..."' simple_name="..." >
            中文
        </xml>'

    Each server entry in the project XML has a minimal XML of the form::

        <KRB_Item>
            <uuid>uuid</uuid>
        </KRB_Item>

    For each server a top-level entry is created::

        <xml item_type="device_server" uuid="..."' simple_name="..." >
            <device_server>
                  <device_instance uuid="..." />
            </device_server>
        </xml>

    Here the uuid in each device_instance refer to device XMLs of the form::

        <xml item_type="device_instance uuid="...">
            <device_instance active_rev="0" instance_id="...">
                <device_config revision="0" uuid="..." />
                <device_config revision="1" uuid="..." />
            </device_instance>
        </xml>

    The mapping dict return as the second tuple element contains the
    `instance_id`s as keys and the `uuid` of the `revision:="0"`
    configurations for each of the 4x4 devices created.

    In total thus 1 Project + 4 x Scene + 4 x Device Server x 4 x Device = 21
    objects are created in the database.

    """
    uuid = _gen_uuid()
    xml = ('<xml item_type="{atype}" uuid="{uuid}" '
           'simple_name="{name}">').format(uuid=uuid, atype='project',
                                           name='Project')
    xml += "<root>"
    xml += "<project>"
    xml += "<scenes>"

    # create some scenes
    for i in range(4):
        sub_uuid = _gen_uuid()
        xml += ('<xml item_type="{atype}" uuid="{uuid}"'
                ' simple_name="{name}" />').format(uuid=sub_uuid,
                                                   atype='scene',
                                                   name=sub_uuid)

        scene_xml = (f'<xml item_type="scene" uuid="{sub_uuid}"'
                     f' simple_name="{sub_uuid}" >'
                     f'<svg:svg xmlns:svg="http://www.w3.org/2000/svg" '
                     f'xmlns:krb="http://karabo.eu/scene" height="768" '
                     f'width="1024" krb:uuid="{sub_uuid}" krb:version="2">'
                     '</svg:svg></xml>')

        db.save_item("LOCAL", sub_uuid, scene_xml)

    xml += "</scenes>"

    # create some device_servers
    xml += "<servers>"
    device_id_conf_map = {}
    for i in range(4):
        sub_uuid = _gen_uuid()

        xml += ('<KRB_Item>'
                '<uuid>{uuid}</uuid>'
                '</KRB_Item>'.format(uuid=sub_uuid))
        ins_xml = ""
        for j in range(4):
            dev_uuid, conf_uuid = create_device(db)
            ins_xml += ('<device_instance '
                        'uuid="{uuid}" />'.format(uuid=dev_uuid))
            device_id_conf_map[dev_uuid] = conf_uuid

        ds_xml = ('<xml item_type="{atype}" uuid="{uuid}"'
                  ' simple_name="{name}" >'
                  '<device_server>'
                  '{instances}'
                  '</device_server>'
                  '</xml>'
                  .format(uuid=sub_uuid, atype='device_server', name=sub_uuid,
                          instances=ins_xml))

        db.save_item("LOCAL", sub_uuid, ds_xml)
    xml += "</servers>"
    xml += "</project>"
    xml += "</root>"
    xml += "</xml>"
    db.save_item("LOCAL", uuid, xml)
    return uuid, device_id_conf_map



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
from os import path as op
from time import strptime

from lxml import etree

from karabo.project_db.bases import DatabaseBase, HandleABC
from karabo.project_db.const import ROOT_COLLECTION, ROOT_COLLECTION_TEST
from karabo.project_db.util import ProjectDBError, to_string

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


class FileHandle(HandleABC):
    def __init__(self, dir):
        self.dir = dir

    def hasCollection(self, path):
        if op.exists(path):
            return True
        else:
            return False

    def removeCollection(self, path):
        """Temporarily do, nothing

        in the long term, we should remove the path"""

    def createCollection(self, path):
        success = False
        try:
            dir_ = op.join(self.dir, path)
            os.makedirs(dir_, exist_ok=True)
            success = True
        except OSError as e:
            raise ProjectDBError(e)
        return success

    def hasDocument(self, path):
        return op.exists(op.join(self.dir, path))

    def load(self, data, path):
        success = False
        try:
            with open(path, 'wb') as f:
                f.write(data)
            success = True
        except OSError as e:
            raise ProjectDBError(e)
        return success

    def getDoc(self, name):
        return self.getDocument(name)

    def getDocument(self, name):
        path = op.join(self.dir, name)
        if not op.exists(path):
            raise ProjectDBError(f"{name} missing")
        with open(path, 'rb') as f:
            return f.read()


class ProjectDatabase(DatabaseBase):

    def __init__(self, dir, test_mode=False, init_db=False):
        """
        Create a project data base context for a given user

        :param dir: the directory where the files are saved
           relative to the '$KARABO'/var/data environment variable
        :param test_mode: defaults to False. In this case
           db_settings.root_collection will be used as the root collection,
           otherwise, root_collection_test is used.
        :param init_db: defaults to False. If True, the default collections
           will be added if missing on context entry.
        :return: a FileDatabase context
        """

        root = (ROOT_COLLECTION if not test_mode else
                ROOT_COLLECTION_TEST)
        self.root = op.join(os.environ['KARABO'], 'var', 'data', dir, root)
        self._init_db = init_db
        self._dbhandle = None

    def path(self, domain, uuid):
        return op.join(self.root, domain, f"{uuid}_0")

    def onEnter(self):
        """Initialize root folder if missing"""
        if self._init_db:
            try:
                os.makedirs(self.root, exist_ok=True)
                os.makedirs(op.join(self.root, "LOCAL"), exist_ok=True)
                self._init_db = False
            except OSError as e:
                raise ProjectDBError(e)
        return FileHandle(self.root)

    def _check_for_modification(self, domain, uuid, old_date):
        """ Check whether the item with of the given `domain` and `uuid` was
        modified

        :param domain: the domain under which this item exists
        :param uuid: the item's uuid
        :param old_date: the item's last modified time stamp
        :return A tuple stating whether the item was modified inbetween and a
                string describing the reason
        """
        # Check whether the object got modified inbetween
        path = self.path(domain, uuid)
        modified = False
        reason = ""
        if op.exists(path):
            tree = etree.parse(path)
            current_date = tree.getroot().attrib['date']
        else:
            # Item not yet existing
            return modified, reason
        current_modified = strptime(current_date, DATE_FORMAT)
        last_modified = strptime(old_date, DATE_FORMAT)
        if last_modified < current_modified:
            # Item got saved inbetween
            modified = True
            reason = "Versioning conflict! Document modified inbetween."
        return modified, reason

    def load_item(self, domain, items):
        """
        Load an item or items from `domain`
        :param domain: a domain to load from
        :param item: the name of the item(s) to load
        :return:
        """
        # path to where the possible entries are located
        assert isinstance(items, (list, tuple))

        results = []
        for uuid in items:
            try:
                paths = (
                    op.join(self.root, domain, f)
                    for f in os.listdir(op.join(self.root, domain))
                    if f.startswith(f'{uuid}_')
                )
                _date = None
                result = None
                for path in paths:
                    r = etree.parse(path).getroot()
                    current_date = r.attrib['date']
                    if not _date or _date < current_date:
                        _date = current_date
                        xml = to_string(r)
                        result = {'uuid': r.attrib['uuid'],
                                  'xml': xml}
                if result:
                    results.append(result)
            except ValueError:
                # empty sequence
                continue
        return results

    def list_items(self, domain, item_types=None):
        """
        List items in domain which match item_types if given, or all items
        if not given
        :param domain: domain to list items from
        :param item_types: list or tuple of item_types to list
        :return: a list of dicts where each entry has keys: uuid, item_type
                 and simple_name
        """
        # path to where the possible entries are located
        dir_ = op.join(self.root, domain)
        results = []
        try:
            for fname in os.listdir(dir_):
                path = op.join(self.root, domain, fname)
                if not op.isfile(path):
                    continue
                r = etree.parse(path).getroot()
                if (item_types is None
                        or r.attrib.get('item_type') in item_types):
                    results.append(
                        {'uuid': r.attrib['uuid'],
                         'item_type': r.attrib.get('item_type'),
                         'simple_name': r.attrib.get('simple_name'),
                         'is_trashed': r.attrib.get('is_trashed', 'false'),
                         'user': r.attrib.get('user', ''),
                         'date': r.attrib.get('date', ''),
                         'description': r.attrib.get('description', '')}
                    )
            return results
        except KeyError as e:
            raise ProjectDBError(e)

    def list_named_items(self, domain, item_type, simple_name):
        """
        List items in domain which match item_type and simple_name

        :param domain: domain to list items from
        :param item_type: item_type to match
        :param simple_name: iterable of names to match
        :return: a list of dicts where each entry has keys: uuid, item_type
                 and simple_name
        """
        # path to where the possible entries are located
        dir_ = op.join(self.root, domain)
        results = []
        try:
            for fname in os.listdir(dir_):
                path = op.join(self.root, domain, fname)
                if not op.isfile(path):
                    continue
                r = etree.parse(path).getroot()
                if (r.attrib.get('simple_name') == simple_name and
                        r.attrib.get('item_type') == item_type):
                    results.append(
                        {'uuid': r.attrib['uuid'],
                         'item_type': r.attrib['item_type'],
                         'simple_name': r.attrib['simple_name'],
                         'is_trashed': r.attrib.get('is_trashed', 'false'),
                         'user': r.attrib.get('user', ''),
                         'date': r.attrib.get('date', ''),
                         'description': r.attrib.get('description', '')}
                    )
            return results
        except KeyError as e:
            raise ProjectDBError(e)

    def list_domains(self):
        """
        List top level domains in database
        :return:
        """
        return [dir_
                for dir_ in os.listdir(self.root)
                if op.isdir(op.join(self.root, dir_))]

    def get_configurations_from_device_name(self, domain, instance_id):
        """Returns a list of configurations for a given device

        :param domain: DB domain
        :param instance_id: instance id of the device
        :return: a list of dicts:
            [{"configid": uuid of the configuration,
              "instanceid": device instance uuid in the DB},
              ...
            ]
        """
        dir_ = op.join(self.root, domain)
        results = []
        for fname in os.listdir(dir_):
            path = op.join(self.root, domain, fname)
            if not op.isfile(path):
                continue
            r = etree.parse(path).getroot()
            if (r.attrib.get('item_type') != 'device_instance'):
                continue
            match = f"device_instance[@instance_id='{instance_id}']"
            inst = r.find(match)
            if inst is not None:
                rev = inst.attrib['active_rev']
                for conf in inst.findall(f"device_config[@revision='{rev}']"):
                    results.append({'instanceid': r.attrib['uuid'],
                                    'configid': conf.attrib['uuid']})
        return results

    def get_projects_from_device(self, domain, uuid):
        """
        Returns the projects which contain a device instance with a given uuid

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a set containing project names
        """

        dir_ = op.join(self.root, domain)
        servers = set()
        for fname in os.listdir(dir_):
            path = op.join(self.root, domain, fname)
            if not op.isfile(path):
                continue
            r = etree.parse(path).getroot()
            if (r.attrib.get('item_type') != 'device_server'):
                continue
            match = f'.//device_instance[@uuid="{uuid}"]'
            if r.findall(match):
                servers.add(r.attrib['uuid'])

        projects = set()

        def add_server(el):
            for server_ in el.findall(".//servers/KRB_Item/uuid"):
                if server_.text in servers:
                    projects.add(el.attrib['simple_name'])
                    return

        for fname in os.listdir(dir_):
            path = op.join(self.root, domain, fname)
            if not op.isfile(path):
                continue
            r = etree.parse(path).getroot()
            if (r.attrib.get('item_type') != 'project'):
                continue
            add_server(r)

        return projects

    def update_attributes(self, items):
        """ Update attribute for the given ``items``

        :param items: list of Hashes containing information on which items
                      to update. Each list entry should be a Hash containing

                      - domain: domain the item resides at
                      - uuid: the uuid of the item
                      - item_type: indicate type of item which attribute should
                                   be changed
                      - attr_name: name of attribute which should be changed
                      - attr_value: value of attribute which should be changed

        :return: a list of dicts where each entry has keys: domain, uuid,
                 item_type, attr_name, attr_value
        """
        res_items = []
        for it in items:
            domain = it.get('domain')
            item_type = it.get('item_type')
            uuid = it.get('uuid')
            attr_name = it.get('attr_name')
            attr_value = it.get('attr_value')
            path = self.path(domain, uuid)
            r = etree.parse(path).getroot()
            if (r.attrib.get('item_type') != item_type):
                continue
            r.attrib[attr_name] = str(attr_value)
            if self.dbhandle.load(to_string(r), path):
                res_items.append(it)
        return res_items



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
from abc import abstractmethod
from asyncio import sleep, wait_for
from uuid import uuid4

from karabo.middlelayer import (
    Device, Hash, String, call, connectDevice, slot, updateDevice)
from karabo.middlelayer.testing import async_tst, sleepUntil
from karabo.project_db.tests.util import create_hierarchy

UUIDS = [str(uuid4()) for i in range(5)]


class ConsumerDevice(Device):
    client = String(
        defaultValue="")

    projectManagerUpdate = String(
        defaultValue="")

    @slot
    async def connectProject(self, instance):
        await self._ss.async_connect(instance, "signalProjectUpdate",
                                     self.slotProject)
        return True

    @slot
    async def slotProject(self, info_hash, deviceId):
        client = info_hash.get("client", "")
        self.client = client
        self.projectManagerUpdate = deviceId
        await sleep(0)


class VerificationProjectManager():
    @abstractmethod
    def _db_init(self):
        """Creates a Database object to inject data in the back-end"""
        raise NotImplementedError

    def stop_local_database(self):
        """Stops the database, overload if needed

        e.g. in the existDb case"""
        pass

    def _createTestData(self):
        with self._db_init() as db:
            path = "{}/{}".format(db.root, 'LOCAL')
            if db.dbhandle.hasCollection(path):
                db.dbhandle.removeCollection(path)

            # make sure we have the LOCAL collection and subcollections
            path = "{}/{}".format(db.root, 'LOCAL')
            if not db.dbhandle.hasCollection(path):
                db.dbhandle.createCollection(path)

            _, self.device_id_conf_map = create_hierarchy(db)

    def _cleanDataBase(self):
        with self._db_init() as db:
            path = "{}/{}".format(db.root, 'LOCAL')
            if db.dbhandle.hasCollection(path):
                db.dbhandle.removeCollection(path)

    def setUp(self):
        # create some test data in the database
        self._createTestData()
        self._ownDir = os.path.dirname(os.path.abspath(__file__))

    def tearDown(self):
        # delete any log files
        # handles both the case where we started as part of integration
        # tests, or as a single test
        dirs = [self._ownDir, os.path.join(self._ownDir, '..')]
        for dir in dirs:
            files = os.listdir(dir)
            for file in files:
                if 'openMQLib.log' in file:
                    os.remove(os.path.join(dir, file))
                if 'device-projManTest' in file:
                    os.remove(os.path.join(dir, file))
                if 'karabo.log' in file:
                    os.remove(os.path.join(dir, file))
                if 'serverId.xml' in file:
                    os.remove(os.path.join(dir, file))

        self._cleanDataBase()
        self.stop_local_database()

    @async_tst
    async def test_project_manager(self):
        # tests are run in sequence as sub tests
        # device server thus is only instantiated once
        # we allow for sleeps in the integration tests as some messaging
        # is async.
        # Additionally, we borrow the SignalSlotable of the device server

        with self.subTest(msg="Test initializing user session"):
            ret = await wait_for(call("projManTest", "slotBeginUserSession",
                                      "admin"), timeout=5)
            self.assertTrue(ret["success"])

        with self.subTest(msg="Test list items"):
            ret = await wait_for(call("projManTest",
                                      "slotListItems", 'admin',
                                      "LOCAL"), timeout=5)
            self.assertTrue(ret.get("success"), ret.get("reason", "no reason"))
            items = ret.get('items')
            self.assertEqual(len(items), 25)
            scenecnt = 0
            for i in items:
                if i.get("item_type") == "scene":
                    scenecnt += 1
            self.assertEqual(scenecnt, 4)

        with self.subTest(msg="Test generic interface"):
            arg = Hash("type", "listItems",
                       "token", "admin",
                       "domain", "LOCAL")
            ret = await wait_for(call("projManTest",
                                      "slotGenericRequest",
                                      arg),
                                 timeout=5)
            self.assertTrue(ret.get("success"), ret.get("reason", "no reason"))
            items = ret.get('items')
            self.assertEqual(len(items), 25)
            scenecnt = 0
            for i in items:
                if i.get("item_type") == "scene":
                    scenecnt += 1
            self.assertEqual(scenecnt, 4)
            # add optional argument. Gets only scenes
            arg["item_types"] = ["scene"]
            ret = await wait_for(call("projManTest",
                                      "slotGenericRequest",
                                      arg),
                                 timeout=5)
            self.assertTrue(ret.get("success"), ret.get("reason", "no reason"))
            items = ret.get('items')
            self.assertEqual(len(items), 4)
            for i in items:
                self.assertEqual(i.get("item_type"), "scene")

        with self.subTest(msg="Test saving data"):
            ret = await wait_for(call("consumeTest", "connectProject",
                                      "projManTest"), timeout=5)
            self.assertTrue(ret)
            proxy = await connectDevice("consumeTest")
            uuid = str(uuid4())
            xml = f'<test uuid="{uuid}">foobar</test>'
            item = Hash()
            item["xml"] = xml
            item["uuid"] = uuid
            item["overwrite"] = False
            # Save a project item!
            item["item_type"] = "project"
            item["domain"] = "LOCAL"
            items = [item, ]
            ret = await wait_for(call("projManTest", "slotSaveItems",
                                      'admin', items, "client-587"), timeout=5)
            items = ret.get("items")
            item = items[0]
            self.assertEqual(item.get('entry.uuid'), uuid)
            self.assertTrue(item.get("success"))

            await updateDevice(proxy)
            await sleepUntil(
                lambda: proxy.projectManagerUpdate == "projManTest",
                timeout=3)
            self.assertEqual(proxy.projectManagerUpdate, "projManTest")
            self.assertEqual(proxy.client, "client-587")

        with self.subTest(msg="Test named list items"):
            # in the project hierarchy created by create_hierarchy
            # there should be only one project and is called "Project"
            ret = await wait_for(call("projManTest",
                                      "slotListNamedItems", 'admin',
                                      "LOCAL", "project",
                                      "Project"), timeout=5)
            items = ret.get('items')
            self.assertEqual(len(items), 1)

        with self.subTest(msg="Test get projects and configurations"):
            for device_id, conf_id in self.device_id_conf_map.items():
                ret = await wait_for(call("projManTest",
                                          "slotListProjectAndConfForDevice",
                                          'admin',
                                          "LOCAL",
                                          device_id,
                                          ), timeout=5)

                self.assertTrue(ret["success"])
                proj = ret["items"][0]["project_name"]
                conf = ret["items"][0]["active_config_ref"]
                self.assertEqual(proj, "Project")
                self.assertEqual(conf_id, conf)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from time import gmtime, strftime, strptime, time

from lxml import etree

from ..bases import DatabaseBase
from ..const import DATE_FORMAT
from ..util import ProjectDBError
from .util import _gen_uuid, create_hierarchy


def create_trashed_project(db, is_trashed=True):
    uuid = _gen_uuid()
    xml = (f'<xml item_type="project" uuid="{uuid}"'
           f' simple_name="{uuid}" is_trashed="{str(is_trashed).lower()}">'
           '</xml>')
    db.save_item("LOCAL", uuid, xml)
    return uuid


def create_unattached_scenes(db):
    # create scenes with the same simple_name
    # unattached to the project
    for i in range(4):
        sub_uuid = _gen_uuid()
        scene_xml = ('<xml item_type="scene" uuid="{uuid}"'
                     ' simple_name="Scene!" >中文</xml>'
                     .format(uuid=sub_uuid))

        db.save_item("LOCAL", sub_uuid, scene_xml)


class ProjectDatabaseVerification():
    """ Base test implementation

    All ProjectDB plugins should implement a test that extends this base
    class.
    """

    def _db_init(self):
        """Should be implemented by the subclass"""

    def stop_local_database(self):
        """Should be implemented by the subclass"""

    def test__make_xml_if_needed(self):
        xml_rep = "<test>foo</test>"
        ret = DatabaseBase._make_xml_if_needed(xml_rep)
        self.assertEqual(ret.tag, "test")
        self.assertEqual(ret.text, "foo")

    def test_malformed_xml_inputs(self):
        xml_rep = "<test>foo</test><bad|symbols></bad|symbols>"
        self.assertRaises(ValueError, DatabaseBase._make_xml_if_needed,
                          xml_rep)

    def test__make_str_if_needed(self):
        element = etree.Element('test')
        element.text = 'foo'
        str_rep = DatabaseBase._make_str_if_needed(element)
        self.assertEqual(str_rep, "<test>foo</test>\n")

    def test_project_interface(self):
        # A bunch of document "names" for the following tests
        testproject = _gen_uuid()
        testproject2 = _gen_uuid()
        with self._db_init() as db:
            # remove previously existing test collection
            path = "{}/{}".format(db.root, 'LOCAL_TEST')
            if db.dbhandle.hasCollection(path):
                db.dbhandle.removeCollection(path)

            path = "{}/{}".format(db.root, 'LOCAL')
            if db.dbhandle.hasCollection(path):
                db.dbhandle.removeCollection(path)

            # make sure we have the LOCAL collection and subcollections
            path = "{}/{}".format(db.root, 'LOCAL')
            if not db.dbhandle.hasCollection(path):
                db.dbhandle.createCollection(path)

            with self.subTest(msg='test_domain_exists'):
                self.assertTrue(db.domain_exists("LOCAL"))

            with self.subTest(msg='test_add_domain'):
                db.add_domain("LOCAL_TEST")
                self.assertTrue(db.domain_exists("LOCAL_TEST"))

            with self.subTest(msg='test_save_item'):
                xml_rep = """
                <xml uuid="{uuid}">foo</xml>
                """.format(uuid=testproject2)

                meta = db.save_item('LOCAL', testproject2, xml_rep)

                path = f"{db.root}/LOCAL/{testproject2}_0"
                self.assertTrue(db.dbhandle.hasDocument(path))
                decoded = db.dbhandle.getDoc(path).decode('utf-8')
                doctree = etree.fromstring(decoded)
                self.assertEqual(doctree.get('uuid'), testproject2)
                self.assertEqual(doctree.text, 'foo')
                self.assertTrue('domain' in meta)
                self.assertTrue('uuid' in meta)

            with self.subTest(msg='test_save_bad_item'):
                xml_rep = """
                <xml uuid="{uuid}"><foo|bar/></xml>
                """.format(uuid=testproject)
                self.assertRaises(ProjectDBError, db.save_item, 'LOCAL',
                                  testproject, xml_rep)

            with self.subTest(msg='load_items'):
                items = [testproject, testproject2]
                res = db.load_item('LOCAL', items)
                for r in res:
                    itemxml = db._make_xml_if_needed(r["xml"])
                    self.assertEqual(itemxml.tag, 'xml')
                    self.assertEqual(itemxml.text, 'foo')

            with self.subTest(msg='test_list_items'):
                create_hierarchy(db)
                items = db.list_items('LOCAL', ['project', 'scene'])
                self.assertEqual(len(items), 5)
                scenecnt = 0
                for i in items:
                    if i["item_type"] == "scene":
                        scenecnt += 1
                    self.assertTrue(i["is_trashed"] == 'false')
                self.assertGreaterEqual(scenecnt, 4)

            with self.subTest(msg='test_trashed_projects'):
                create_trashed_project(db)
                items = db.list_items('LOCAL', ['project'])
                self.assertEqual(len(items), 2)
                nb_is_trashed = 0
                for it in items:
                    if it["is_trashed"] == 'true':
                        nb_is_trashed += 1
                self.assertEqual(nb_is_trashed, 1)

            with self.subTest(msg='test_list_domains'):
                items = db.list_domains()
                for it in items:
                    self.assertTrue(it in ['LOCAL_TEST', 'REPO', 'LOCAL'])

            with self.subTest(msg='test_old_versioned_data'):
                versioned_uuid = _gen_uuid()
                xml_rep = """
                <xml uuid="{uuid}" revision="{revision}" date="{date}"
                     user="sue">foo</xml>
                """
                MAX_REV = 3
                earlier = time() - 5.0
                # Save three consecutive versions
                for i in range(MAX_REV, 0, -1):
                    # consecutive dates beginning a few seconds ago
                    date = strftime(DATE_FORMAT, gmtime(earlier + i))
                    xml = xml_rep.format(uuid=versioned_uuid, revision=i,
                                         date=date)
                    path = "{}/{}/{}_{}".format(db.root, 'LOCAL',
                                                versioned_uuid, i)
                    # Save directly without engaging save_item()
                    db.dbhandle.load(xml.encode('utf-8'), path)

                # Read it back, no revision given
                res = db.load_item('LOCAL', [versioned_uuid])
                self.assertEqual(len(res), 1)
                itemxml = db._make_xml_if_needed(res[0]['xml'])
                self.assertEqual(int(itemxml.get('revision')), MAX_REV)

                # Write it back with save_item
                date = strftime(DATE_FORMAT, gmtime())
                xml_rep = """
                <xml uuid="{uuid}" date="{date}" user="sue">foo</xml>
                """.format(uuid=versioned_uuid, date=date)
                meta = db.save_item('LOCAL', versioned_uuid, xml_rep)

                # Make sure there's a revision of 0 (auto-added)
                res = db.load_item('LOCAL', [versioned_uuid])
                itemxml = db._make_xml_if_needed(res[0]['xml'])
                self.assertEqual(itemxml.get('revision'), '0')

            with self.subTest(msg='test_named_items'):
                create_unattached_scenes(db)
                items = db.list_named_items('LOCAL', 'scene', 'Scene!')
                self.assertEqual(len(items), 4)
                scenecnt = 0
                for i in items:
                    if i["item_type"] == "scene":
                        scenecnt += 1
                    self.assertEqual(i["simple_name"], "Scene!")
                self.assertGreaterEqual(scenecnt, 4)

            with self.subTest(msg='test_get_projects_and_configs'):
                _, device_id_conf_map = create_hierarchy(db)
                for device_id, conf_id in device_id_conf_map.items():
                    items = db.get_projects_with_conf("LOCAL", device_id)
                    proj, conf = next(iter(items.items()))
                    self.assertEqual(proj, "Project")
                    self.assertEqual(conf_id, conf)

            self.stop_local_database()

    def test_save_check_modification(self):
        proj_uuid = _gen_uuid()

        with self._db_init() as db:

            path = "{}/{}".format(db.root, 'LOCAL')
            if db.dbhandle.hasCollection(path):
                db.dbhandle.removeCollection(path)

            # make sure we have the LOCAL collection and subcollections
            path = "{}/{}".format(db.root, 'LOCAL')
            if not db.dbhandle.hasCollection(path):
                db.dbhandle.createCollection(path)

            with self.subTest(msg='test_save_item'):
                date = "2011-11-01 09:00:52"
                xml_rep = """
                <xml uuid="{uuid}" date="{date}">foo</xml>
                """.format(uuid=proj_uuid, date=date)

                success = True
                try:
                    meta = db.save_item('LOCAL', proj_uuid, xml_rep)
                except ProjectDBError:
                    success = False

                self.assertTrue(success)
                self.assertTrue('date' in meta)
                new_date = meta['date']
                self.assertNotEqual(new_date, date)
                date_t = strptime(date, DATE_FORMAT)
                current_date_t = strptime(new_date, DATE_FORMAT)
                self.assertLess(date_t, current_date_t)

                # Try saving again but using same time stamp
                success = True
                reason = ""
                try:
                    meta = db.save_item('LOCAL', proj_uuid, xml_rep)
                except ProjectDBError as e:
                    success = False
                    reason = str(e)
                self.assertFalse(success)
                self.assertTrue(reason)

            self.stop_local_database()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import inspect
import logging
import os
import socket
import time
import traceback
import weakref
from asyncio import (
    CancelledError, Event, Future, Lock, TimeoutError, current_task,
    ensure_future, gather, get_event_loop, iscoroutinefunction, shield, sleep,
    wait_for)
from contextlib import AsyncExitStack
from functools import partial, wraps
from itertools import count

import aiormq
import numpy
from aiormq.base import TaskWrapper

from karabo.native import (
    Hash, KaraboError, decodeBinary, decodeBinaryPos, encodeBinary)

from .utils import suppress

_OVERFLOW_POLICY = "drop-head"  # Drop oldest messages
_TIME_TO_LIVE = 120_000  # 120 seconds expiry time [ms]
_BEATS_QSIZE = 12_000
_CONSUME_QSIZE = 100_000


class KaraboWrapper(TaskWrapper):
    """The purpose of this class is to flag external tasks"""

    def __init__(self, task) -> None:
        super().__init__(task)
        task.shielded = True


aiormq.base.TaskWrapper = KaraboWrapper


def ensure_running(func):
    """Ensure a running eventloop for `func`"""
    if iscoroutinefunction(func):
        async def wrapper(self, *args, **kwargs):
            if self.loop.is_closed() or not self.loop.is_running():
                self.logger.warning(
                    f"Loop not running for func {func.__name__} with "
                    f"arguments {args}.")
                return
            return await func(self, *args, **kwargs)
    else:
        def wrapper(self, *args, **kwargs):
            if self.loop.is_closed() or not self.loop.is_running():
                self.logger.warning(
                    f"Loop not running for func {func.__name__} with "
                    f"arguments {args}.")
                return
            return func(self, *args, **kwargs)
    return wrapper


class Broker:
    def __init__(self, loop, deviceId, classId, broadcast=True):
        self.domain = loop.topic
        self.loop = loop
        self.connection = None
        self.deviceId = deviceId
        self.classId = classId
        self.brokerId = None
        # Interest in receiving broadcasts
        self.broadcast = broadcast
        self.logger = logging.getLogger(deviceId)
        # Basics
        self.exitStack = AsyncExitStack()
        self.slots = {}
        self.info = None
        self.repliers = {}
        self.tasks = set()
        # AMQP specific bookkeeping ...
        self.subscriptions = set()  # registered tuple: exchange, routing key
        self.channel = None  # channel for communication
        self.queue = None  # main queue
        self.consumer_tag = None  # tag returned by consume method
        self.heartbeat_queue = None
        self.heartbeat_consumer_tag = None
        self.exit_event = Event()
        self.heartbeat_task = None
        self.subscribe_lock = Lock()
        self.hostname = socket.gethostname()
        # Flag to indicate when a channel is about to be closed
        self.shutdown_channel = False

    async def subscribe_default(self):
        """Subscribe to 'default' exchanges to allow a communication
        through the broker:
            -------------
            exchange = <domain>.global_slots
            routing_key = ""
            queue = <deviceId>
            -------------
            exchange = <domain>.slots
            routing_key = <deviceId>
            queue = <deviceId>
        """
        self.brokerId = f"{self.domain}.{self.deviceId}"
        try:
            await self.channel.queue_declare(self.brokerId, passive=True)
            # If no exception raised the the queue name exists already ...
            # To continue  just use generated queue name...
            timestamp = hex(int(time.monotonic() * 1000000000))[2:]
            self.brokerId = f"{self.domain}.{self.deviceId}:{timestamp}"
        except aiormq.exceptions.ChannelNotFoundEntity:
            # Exception raised since the queue not found on the broker
            # The channel is not valid anymore, so create the new one
            self.channel = await self.connection.channel(
                publisher_confirms=False)

        arguments = {
            "x-max-length": _CONSUME_QSIZE,
            "x-overflow": _OVERFLOW_POLICY,
            "x-message-ttl": _TIME_TO_LIVE}
        declare_ok = await self.channel.queue_declare(
            self.brokerId, auto_delete=True, arguments=arguments)
        # The received queue name (either input or generated) is ...
        self.queue = declare_ok.queue
        # create main exchange : <domain>.slots
        exchange = f"{self.domain}.slots"
        await self.channel.exchange_declare(exchange=exchange,
                                            exchange_type="topic")
        key = self.deviceId
        await self.channel.queue_bind(self.queue, exchange, routing_key=key)
        async with self.subscribe_lock:
            self.subscriptions.add((exchange, key))

        # Globals for instanceNew, Gone ...
        exchange = f"{self.domain}.global_slots"
        await self.channel.exchange_declare(exchange=exchange,
                                            exchange_type="topic")
        # Only if interested, we subscribe to broadcasts messages,
        # but still declare the exchange to publish
        if self.broadcast:
            key = ""
            await self.channel.queue_bind(self.queue, exchange,
                                          routing_key=key)
            async with self.subscribe_lock:
                self.subscriptions.add((exchange, key))

        exchange = f"{self.domain}.signals"
        await self.channel.exchange_declare(exchange=exchange,
                                            exchange_type="topic")

    def enter_context(self, context):
        """Synchronously enter the exit stack context"""
        return self.exitStack.enter_context(context)

    async def enter_async_context(self, context):
        """Asynchronously enter the exit stack context"""
        return await self.exitStack.enter_async_context(context)

    async def main(self, device):
        """This is the main loop of a device (SignalSlotable instance)

        A device is running if this coroutine is still running.
        Use `stop_tasks` to stop this main loop."""
        async with self.exitStack:
            device = weakref.ref(device)
            await self.consume(device())

    def register_slot(self, name, slot):
        """register a slot on the device

        :param name: the name of the slot
        :param slot: the slot to be called. If this is a bound method, it is
            assured that no reference to the object holding the method is kept.
        """
        if inspect.ismethod(slot):
            def delete(ref):
                del self.slots[name]

            weakself = weakref.ref(slot.__self__, delete)
            func = slot.__func__

            if iscoroutinefunction(func):
                @wraps(func)
                async def wrapper(*args):
                    return await func(weakself(), *args)
            else:
                @wraps(func)
                def wrapper(*args):
                    return func(weakself(), *args)

            self.slots[name] = wrapper
        else:
            self.slots[name] = slot

    def updateInstanceInfo(self, info):
        """update the short information about this instance

        the instance info hash contains a very brief summary of the device.
        It is regularly published, and even lives longer than a device,
        as it is published with the message that the device died."""
        self.info.merge(info)
        self.emit("call", {"*": ["slotInstanceUpdated"]},
                  self.deviceId, self.info)

    def replyException(self, message, exception):
        trace = ''.join(traceback.format_exception(
            type(exception), exception, exception.__traceback__))
        self.reply(message, (str(exception), trace), error=True)

    def emit(self, signal, targets, *args):
        self.call(signal, targets, None, args)

    async def request(self, device, target, *arguments):
        reply = f"{self.deviceId}-{time.monotonic().hex()[4:-4]}"
        self.call("call", {device: [target]}, reply, arguments)
        future = Future(loop=self.loop)
        self.repliers[reply] = future
        future.add_done_callback(lambda _: self.repliers.pop(reply))
        return (await future)

    def encode_binary_message(self, header, arguments):
        body = Hash()
        for i, a in enumerate(arguments):
            body[f"a{i + 1}"] = a
        # Add timestamp (epoch in ms) when message is sent
        header["MQTimestamp"] = numpy.int64(time.time() * 1000)
        header["signalInstanceId"] = self.deviceId
        return encodeBinary(header) + encodeBinary(body)

    @ensure_running
    def send(self, exch, key, header, arguments):
        msg = self.encode_binary_message(header, arguments)
        ensure_future(self.channel.basic_publish(
            msg, routing_key=key, exchange=exch))

    async def async_send(self, exch, key, header, arguments):
        msg = self.encode_binary_message(header, arguments)
        await self.channel.basic_publish(msg, routing_key=key, exchange=exch)

    async def heartbeat(self, interval, info):
        header = Hash("signalFunction", "signalHeartbeat")
        header["signalInstanceId"] = self.deviceId
        body = Hash(
            "a1", self.deviceId,
            "a2", interval,
            "a3", info)
        msg = encodeBinary(header) + encodeBinary(body)
        exch = f"{self.domain}.signals"
        key = f"{self.deviceId}.signalHeartbeat"
        await self.channel.basic_publish(msg, routing_key=key, exchange=exch)

    async def notify_network(self, info):
        """notify the network that we are alive

        we send out an instance new and gone, and the heartbeats in between.

        :param info: the info Hash that should be published regularly.
        """
        self.info = info

        interval = self.info["heartbeatInterval"]
        heartbeat_info = Hash("type", info["type"],
                              "heartbeatInterval", interval)
        await self.async_emit("call", {"*": ["slotInstanceNew"]},
                              self.deviceId, self.info)

        async def heartbeat():
            try:
                while True:
                    await sleep(interval)
                    await self.heartbeat(interval, heartbeat_info)
            except CancelledError:
                pass
            finally:
                await self.async_emit("call", {"*": ["slotInstanceGone"]},
                                      self.deviceId, self.info)

        self.heartbeat_task = ensure_future(heartbeat())

    def build_arguments(self, signal, targets, reply):
        funcs = "||".join(f"{k}:{','.join(v)}" for k, v in targets.items())
        slotInstanceIds = "||".join(targets)
        p = Hash(
            "signalFunction", signal,
            "slotInstanceIds", f"|{slotInstanceIds}|",
            "slotFunctions", f"|{funcs}|",
            "hostname", self.hostname,
            "classId", self.classId)
        if reply is not None:
            p.setElement("replyTo", reply, {})

        # AMQP specific follows ...
        if signal in {"__replyNoWait__", "__reply__"}:
            name = f"{self.domain}.slots"
            routing_key = slotInstanceIds
        elif signal == "call":
            if slotInstanceIds == "*":
                name = f"{self.domain}.global_slots"
                routing_key = ""
            else:
                name = f"{self.domain}.slots"
                routing_key = slotInstanceIds
        else:
            name = f"{self.domain}.signals"
            routing_key = f"{self.deviceId}.{signal}"

        return name, routing_key, p

    def call(self, signal, targets, reply, arguments):
        if not targets:
            return
        name, routing_key, p = self.build_arguments(signal, targets, reply)
        self.send(name, routing_key, p, arguments)

    async def async_call(self, signal, targets, reply, arguments):
        if not targets:
            return
        name, routing_key, p = self.build_arguments(signal, targets, reply)
        await self.async_send(name, routing_key, p, arguments)

    async def async_emit(self, signal, targets, *arguments):
        await self.async_call(signal, targets, None, arguments)

    def reply(self, message, reply, error=False):
        header = message["header"]
        sender = header["signalInstanceId"]

        if not isinstance(reply, tuple):
            reply = reply,

        if replyTo := header.get("replyTo"):
            p = Hash(
                "replyFrom", replyTo,
                "signalFunction", "__reply__",
                "slotInstanceIds", f"|{sender}|",
                "error", error)
            name = f"{self.domain}.slots"
            routing_key = sender
            self.send(name, routing_key, p, reply)

        if replyId := header.get("replyInstanceIds"):
            p = Hash(
                "signalFunction", "__replyNoWait__",
                "slotInstanceIds", replyId,
                "slotFunctions", header["replyFunctions"],
                "error", error)
            dest = replyId.strip("|")
            name = f"{self.domain}.slots"
            routing_key = dest
            self.send(name, routing_key, p, reply)

    def connect(self, deviceId, signal, slot):
        """This is way of establishing 'karabo signalslot' connection with
        a signal.  In AMQP case this is just creating a task that will
        call `async_connect` that, in turn, may take time because communication
        with RabbitMQ broker is involved.
        """
        self.loop.call_soon_threadsafe(
            self.loop.create_task, self.async_connect(deviceId, signal, slot))

    async def async_connect(self, deviceId, signal, slot):
        """This is way of establishing karabo connection between local slot and
        remote signal.  In AMQP case this is two steps procedure:
        1. subscription to the topic with topic name built from signal info
        2. sending to the signal side the slot registration message
        NOTE: Optimization: we can connect many signals to the same alot
        at once
        """
        if isinstance(signal, (list, tuple)):
            signals = list(signal)
        else:
            signals = [signal]

        exchange = f"{self.domain}.signals"

        futures = [sleep(0)]
        async with self.subscribe_lock:
            for s in signals:
                key = f"{deviceId}.{s}"
                subscription = (exchange, key)
                if subscription not in self.subscriptions:
                    futures.append(self.channel.queue_bind(
                        queue=self.queue, exchange=exchange, routing_key=key))
                    self.subscriptions.add(subscription)

                futures.append(
                    self.async_emit(
                        "call", {deviceId: ["slotConnectToSignal"]},
                        s, slot.__self__.deviceId, slot.__name__))

        await gather(*futures, return_exceptions=True)

    @ensure_running
    def disconnect(self, devId, signal, slot):
        self.loop.call_soon_threadsafe(
            self.loop.create_task, self.async_disconnect(devId, signal, slot))

    @ensure_running
    async def async_disconnect(self, deviceId, signal, slot):
        signals = []
        if isinstance(signal, (list, tuple)):
            signals = list(signal)
        else:
            signals = [signal]

        exchange = f"{self.domain}.signals"
        try:
            futures = [sleep(0)]
            async with self.subscribe_lock:
                for s in signals:
                    key = f"{deviceId}.{s}"
                    subscription = (exchange, key)
                    if subscription in self.subscriptions:
                        self.subscriptions.remove(subscription)

                        futures.append(self.channel.queue_unbind(
                            queue=self.queue, exchange=exchange,
                            routing_key=key))
                    futures.append(self.async_emit(
                        "call", {deviceId: ["slotDisconnectFromSignal"]},
                        s, slot.__self__.deviceId, slot.__name__))
            await gather(*futures, return_exceptions=True)
        except BaseException:
            self.logger.warning(
                f"Fail to disconnect from signals: {signals}")

    async def async_unsubscribe_all(self):
        futures = [sleep(0)]
        async with self.subscribe_lock:
            futures.extend([
                shield(self.channel.queue_unbind(
                    queue=self.queue, exchange=exchange, routing_key=key)
                ) for exchange, key in self.subscriptions])
            self.subscriptions = set()
        await gather(*futures, return_exceptions=True)

    async def handleMessage(self, message, device):
        """Decode message from binary blob
        has valid information to do so...  Otherwise
        simply call handle the message as usual...
        """
        try:
            header, pos = decodeBinaryPos(message)
            body = decodeBinary(message[pos:])
            decoded = Hash("header", header, "body", body)
        except BaseException:
            self.logger.exception("Malformed message")
            return

        try:
            slots, params = self.decodeMessage(decoded)
        except BaseException:
            self.logger.exception("Malformed message")
            return
        try:
            callSlots = [(self.slots[s], s)
                         for s in slots.get(self.deviceId, [])]
            if self.broadcast:
                callSlots.extend(
                    [(self.slots[s], s)
                     for s in slots.get("*", []) if s in self.slots])
        except KeyError as e:
            text = f"Slot does not exist: {e}"
            self.logger.error(text)
            exc = KaraboError(text).with_traceback(e.__traceback__)
            self.replyException(decoded, exc)
            return
        try:
            for slot, name in callSlots:
                # call "coslot.outer(...)" for slot, i.e. reply
                slot.slot(slot, device, name, decoded, params)
        except Exception as e:
            # the slot.slot wrapper should already catch all exceptions
            # all exceptions raised additionally are a bug in Karabo
            text = f"Internal error while executing slot: {e}"
            self.logger.error(text)
            exc = KaraboError(text).with_traceback(e.__traceback__)
            self.replyException(decoded, exc)

    async def _stop_heartbeat(self):
        if self.heartbeat_task is not None:
            if not self.heartbeat_task.done():
                self.heartbeat_task.cancel()
                await self.heartbeat_task
            self.heartbeat_task = None

    async def stop_tasks(self):
        """Reimplemented method of `Broker`"""
        await self._on_stop_tasks()
        me = current_task(loop=None)
        # Services that are listening to broadcasts don't need their
        # channel closed, e.g. clients or servers. This will happen
        # on connection closure
        if not self.broadcast:
            me.add_done_callback(self.close_channel)

        tasks = [t for t in self.tasks if t is not me
                 and not getattr(t, "shielded", False)]
        for t in tasks:
            t.cancel()
        try:
            await wait_for(gather(*tasks, return_exceptions=True),
                           timeout=5)
            return True
        except TimeoutError:
            return False

    def close_channel(self, cb):
        """Close channel callback after slotKillDevice"""
        if self.shutdown_channel:
            return
        self.shutdown_channel = True

        async def closing_channel():
            if self.channel is not None:
                await self.channel.close()
                self.channel = None
            self.connection = None

        ensure_future(closing_channel())

    async def _on_stop_tasks(self):
        if self.consumer_tag is not None:
            with suppress(BaseException):
                await self.channel.basic_cancel(self.consumer_tag)
            self.consumer_tag = None
        if self.heartbeat_consumer_tag is not None:
            with suppress(BaseException):
                await self.channel.basic_cancel(self.heartbeat_consumer_tag)
            self.heartbeat_consumer_tag = None
        with suppress(BaseException):
            await self.exitStack.aclose()
        self.exit_event.set()
        await self._stop_heartbeat()
        await self.async_unsubscribe_all()

    async def on_message(self, device, message):
        d = device()
        if d is not None:
            await self.handleMessage(message.body, d)

    async def consume(self, device):
        device = weakref.ref(device)
        consume_ok = await self.channel.basic_consume(
            self.queue, partial(self.on_message, device),
            exclusive=True, no_ack=True)
        # no_ack means automatic acknowlegdement
        self.consumer_tag = consume_ok.consumer_tag
        # Be under exitStack scope as soon as queue is alive
        await self.exit_event.wait()

    async def on_heartbeat(self, device, message):
        d = device()
        if d is not None:
            message = message.body
            try:
                _, pos = decodeBinaryPos(message)
                hsh = decodeBinary(message[pos:])
            except BaseException:
                self.logger.exception("Malformed message")
            else:
                instance_id, info = hsh["a1"], hsh["a3"]
                await d.updateHeartBeat(instance_id, info)

    async def consume_beats(self, device):
        """Heartbeat method for the device server"""
        device = weakref.ref(device)
        name = f"{self.brokerId}:beats"
        arguments = {
            "x-max-length": _BEATS_QSIZE,
            "x-overflow": _OVERFLOW_POLICY,
            "x-message-ttl": _TIME_TO_LIVE}
        declare_ok = await self.channel.queue_declare(
            name, auto_delete=True, arguments=arguments)
        self.heartbeat_queue = declare_ok.queue
        consume_ok = await self.channel.basic_consume(
            self.heartbeat_queue, partial(self.on_heartbeat, device),
            no_ack=True)
        self.heartbeat_consumer_tag = consume_ok.consumer_tag

        # Binding and book-keeping!
        exchange = f"{self.domain}.signals"
        key = "*.signalHeartbeat"
        await self.channel.queue_bind(self.heartbeat_queue, exchange,
                                      routing_key=key)
        async with self.subscribe_lock:
            self.subscriptions.add((exchange, key))

    def decodeMessage(self, hash):
        """Decode a Karabo message

        reply messages are dispatched directly.

        :returns: a dictionary that maps the device id of slots to be called
            to a list of slots to be called on that device
        """
        header = hash["header"]
        body = hash["body"]
        params = []
        for i in count(1):
            try:
                params.append(body[f"a{i}"])
            except KeyError:
                break

        replyFrom = header.get("replyFrom")
        if replyFrom is not None:
            f = self.repliers.get(replyFrom)
            if f is not None and not f.done():
                if header.get("error", False):
                    exceptTxt = params[0]
                    if len(params) >= 2 and params[1]:
                        exceptTxt += "\nDETAILS: " + params[1]
                    f.set_exception(KaraboError(exceptTxt))
                else:
                    if len(params) == 1:
                        params = params[0]
                    else:
                        params = tuple(params)
                    f.set_result(params)
            return {}, None

        slots = (header["slotFunctions"][1:-1]).split("||")
        return ({k: v.split(",") for k, v in (s.split(":") for s in slots)},
                params)

    def get_property(self, message, prop):
        header = message.get("header")
        return header.get(prop) if header is not None else None

    async def create_global_connection(self):
        loop = get_event_loop().global_loop
        connection = loop.connection
        if connection and connection.is_opened:
            return connection
        urls = os.environ.get("KARABO_BROKER",
                              "amqp://localhost:5672").split(",")
        for url in urls:
            try:
                # Perform connection
                connection = await aiormq.connect(url)
                loop.connection = connection
                break
            except BaseException as e:
                print(f"While trying node '{url}': {str(e)}")
                connection = None
        return connection

    async def ensure_connection(self):
        if self.connection is None:
            self.connection = await shield(
                self.create_global_connection())
        if self.connection is None:
            raise RuntimeError("No connection established")
        # Creating a channel
        self.channel = await self.connection.channel(
            publisher_confirms=False)
        await self.channel.basic_qos(prefetch_count=1)
        await self.subscribe_default()

    @staticmethod
    def create_connection(hosts, connection):
        return connection



#############################################################################
# Author: degon & costar
# Created on August 21, 2020, 11:42 AM
# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os
import os.path as op
import re
from asyncio import CancelledError, Future, TimeoutError, gather, wait_for
from collections import defaultdict

from karabo.common.scenemodel.api import (
    BoxLayoutModel, DisplayCommandModel, DisplayLabelModel,
    DisplayStateColorModel, DisplayTextLogModel, EditableRegexModel,
    ErrorBoolModel, IntLineEditModel, LabelModel, LineEditModel, LineModel,
    SceneModel, TableElementModel, write_scene)
from karabo.config_db import (
    ConfigurationDatabase, DbHandle, hashFromBase64Bin, hashToBase64Bin,
    schemaFromBase64Bin, schemaToBase64Bin)
from karabo.middlelayer import (
    AccessLevel, AccessMode, Assignment, Bool, Configurable, DaqPolicy,
    DeviceClientBase, Hash, HashList, KaraboError, Overwrite, RegexString,
    Slot, State, String, Timestamp, UInt32, VectorHash, VectorString,
    background, dictToHash, sanitize_init_configuration, slot)

HIDDEN_KARABO_FOLDER = op.join(os.environ['HOME'], '.karabo')

DEVICE_TIMEOUT = 3
INSTANCE_NEW_TIMEOUT = 15
FILTER_KEYS = ["name", "timepoint", "description", "priority"]


def scratch_conf(c: dict) -> dict:
    """Create a scoped down dictionary of the configuration"""
    ret = {k: v for k, v in c.items() if k in FILTER_KEYS}
    return ret


class RowSchema(Configurable):
    name = String(
        defaultValue="",
        description="The name of the configuration",
        accessMode=AccessMode.READONLY)

    timepoint = String(
        defaultValue="",
        description="The timepoint when the configuration was saved",
        accessMode=AccessMode.READONLY)

    description = String(
        defaultValue="",
        description="The description of the configuration",
        accessMode=AccessMode.READONLY)

    priority = UInt32(
        defaultValue=1,
        description="The priority of the configuration",
        accessMode=AccessMode.READONLY)


class ConfigurationManager(DeviceClientBase):
    """This configuration manager service is to control device configurations

    Requests from various clients (GUI's and devices) will be managed.

    - Tag device configurations by ``name``
    - Retrieve device configurations by ``name``
    - List device configurations by ``deviceId``
    """

    state = Overwrite(
        defaultValue=State.INIT,
        options=[State.INIT, State.CHANGING, State.UNKNOWN, State.ON])

    visibility = Overwrite(
        defaultValue=AccessLevel.ADMIN,
        options=[AccessLevel.ADMIN])

    @String(
        defaultValue="",
        description="The deviceId to look for a configuration",
        requiredAccessLevel=AccessLevel.OPERATOR)
    async def deviceName(self, value):
        if value is not None:
            self.deviceName = value
            if self.db is not None:
                background(self._list_configurations())

    configurationName = RegexString(
        regex="^[A-Za-z0-9_-]{1,30}$",
        defaultValue="default",
        description="The configuration name",
        requiredAccessLevel=AccessLevel.OPERATOR)

    priority = UInt32(
        defaultValue=1,
        minInc=1,
        maxInc=3,
        description="The priority of the configuration",
        requiredAccessLevel=AccessLevel.OPERATOR)

    description = String(
        defaultValue="",
        description="The configuration description",
        requiredAccessLevel=AccessLevel.OPERATOR)

    availableScenes = VectorString(
        displayedName="Available Scenes",
        displayType="Scenes",
        description="Provides a scene for the Configuration Manager.",
        accessMode=AccessMode.READONLY,
        daqPolicy=DaqPolicy.OMIT,
        defaultValue=['scene'])

    lastSuccess = Bool(
        defaultValue=True,
        displayedName="Last Success",
        description="Indicates if the last action was successful",
        accessMode=AccessMode.READONLY)

    view = VectorHash(
        rows=RowSchema,
        defaultValue=[],
        displayedName="View",
        accessMode=AccessMode.READONLY,
        daqPolicy=DaqPolicy.OMIT)

    dbName = String(
        defaultValue="karaboDB",
        displayedName="DB Name",
        description="The database name",
        requiredAccessLevel=AccessLevel.OPERATOR,
        assignment=Assignment.MANDATORY,
        accessMode=AccessMode.INITONLY)

    @slot
    def requestScene(self, params):
        """Fulfill a scene request from another device.

        :param params: A `Hash` containing the method parameters
        """
        payload = Hash('success', False)
        name = params.get('name', default='scene')
        if name == 'scene':
            payload.set('success', True)
            payload.set('name', name)
            payload.set('data', get_scene(self.deviceId))

        return Hash('type', 'deviceScene',
                    'origin', self.deviceId,
                    'payload', payload)

    @Slot(displayedName="List configurations",
          requiredAccessLevel=AccessLevel.OPERATOR,
          allowedStates=[State.ON])
    async def listConfigurations(self):
        """List all configuration for a device in the table element!"""
        self.state = State.CHANGING
        background(self._list_configurations())

    async def _list_configurations(self):
        deviceId = self.deviceName.value
        name_part = ""
        items = []
        try:
            items = self.db.list_configurations(deviceId, name_part)
            # Adjust for the table element!
            items = [scratch_conf(c) for c in items]
            # Convert to a list of Hashes!
            items = [dictToHash(c) for c in items]
        except Exception as e:
            self.status = str(e)
            self.lastSuccess = False
        else:
            self.status = f"Listing configurations for {deviceId}"
            self.lastSuccess = True

        self.view = items
        self.state = State.ON

    @Slot(displayedName="Save configurations",
          requiredAccessLevel=AccessLevel.OPERATOR,
          allowedStates=[State.ON])
    async def saveConfigurations(self):
        """Save a configuration for a given `deviceId`"""
        self.state = State.CHANGING
        background(self._save_configuration())

    async def _save_configuration(self):
        deviceId = self.deviceName.value
        priority = int(self.priority.value)
        description = self.description.value
        config_name = self.configurationName.value

        taken = self.db.is_config_name_taken(config_name, [deviceId])
        if taken:
            self.status = (f"Failure: Configuration name {config_name} "
                           f"is already taken for device {deviceId}")
            self.lastSuccess = False
            self.state = State.ON
            return
        # Get the schema and the configuration first!
        try:
            schema, _ = await wait_for(self.call(
                deviceId, "slotGetSchema", False), timeout=DEVICE_TIMEOUT)
            conf, _ = await wait_for(self.call(
                deviceId, "slotGetConfiguration"), timeout=DEVICE_TIMEOUT)
        except (CancelledError, TimeoutError):
            self.status = (f"Failure: Saving configuration for {deviceId} "
                           f"failed. The device is not online.")
            self.lastSuccess = False
        except Exception as e:
            # Handle unexpected error
            self.status = str(e)
            self.lastSuccess = False
        else:
            configs = {}
            configs.update({"deviceId": deviceId})
            configs.update({"config": hashToBase64Bin(conf)})
            configs.update({"schema": schemaToBase64Bin(schema)})
            items = [configs]
            # Now we save and list again, and we should not expect any errors
            try:
                self.db.save_configuration(
                    config_name, items, description=description, user=".",
                    priority=priority, timestamp="")

                self.status = (f"Success: Saved configuration {config_name} "
                               f"for device {deviceId}!")
                current_items = self.db.list_configurations(
                    deviceId, name_part="")
                current_items = [scratch_conf(c) for c in current_items]
                current_items = [dictToHash(c) for c in current_items]
                self.view = current_items
            except Exception as e:
                # Handle unexpected errors as bulk
                self.status = str(e)
                self.lastSuccess = False
            else:
                self.lastSuccess = True

        # At the very end, go to `ON` state again
        self.state = State.ON

    confBulkLimit = UInt32(
        defaultValue=30,
        description="The limit of configurations allowed in a single save "
                    "configuration call",
        requiredAccessLevel=AccessLevel.ADMIN,
        accessMode=AccessMode.READONLY)

    @slot
    async def slotInstanceNew(self, instanceId, info):
        if info["type"] == "server":
            self._class_schemas.pop(instanceId, None)
        await super().slotInstanceNew(instanceId, info)

    @slot
    def slotInstanceGone(self, instanceId, info):
        if info["type"] == "server":
            self._class_schemas.pop(instanceId, None)
        return super().slotInstanceGone(instanceId, info)

    def __init__(self, configuration):
        super().__init__(configuration)
        self.db = None
        regex = self.configurationName.descriptor.regex
        self.name_pattern = re.compile(regex)

        # Dictionary of serverId: {classId: schema}
        self._class_schemas = defaultdict(dict)

    async def onInitialization(self):
        """Initialize the configuration database device and create the `DB`"""
        folder = op.join(os.environ["KARABO"], 'var', 'data', 'config_db')
        path = op.join(folder, self.dbName.value)
        dir_name = op.dirname(path)
        if not op.exists(dir_name):
            os.makedirs(dir_name, exist_ok=True)

        connection = DbHandle(path)
        if not connection.authenticate():
            # XXX: We bring the device down if the db is not available!
            raise KaraboError(f"Database with address {self.dbName.value} is"
                              f"not available.")
        self.db = ConfigurationDatabase(connection)
        self.db.assureExisting()
        self.state = State.ON

        self._schema_futures = {}
        # If we have a device name already we can retrieve a list!
        if self.deviceName:
            background(self._list_configurations())

        self.logger.info(f"Starting configuration manager {self.deviceId}")

    # Karabo Slot Interface
    # -----------------------------------------------------------------------

    @slot
    async def slotListConfigurationFromName(self, info):
        """Slot to list configurations from name

        This slot requires an info Hash with `deviceId`
        """
        deviceId = info["deviceId"]
        name_part = info.get("name", "")

        items = self.db.list_configurations(deviceId, name_part)
        items = HashList([dictToHash(config) for config in items])

        return Hash("success", True, "items", items)

    @slot
    async def slotListConfigurationSets(self, info):
        """Slot to list configurations from name

        This slot requires an info Hash with:
            - `deviceIds`: list of deviceIds whose configurations should be
                           listed.

        :returns: list of configuration items (can be empty) with meta
                  information.
                  A configuration item has keys `name`, `priority`,
                  `description`, `user`, `min_point`, `max_timepoint`
                  and `diff_timepoint`.
        """
        deviceIds = info["deviceIds"]

        items = self.db.list_configuration_sets(deviceIds)
        items = HashList([dictToHash(config) for config in items])

        return Hash("success", True, "items", items)

    @slot
    async def slotGetConfigurationFromName(self, info):
        """Slot to get a configuration from name

        The info `Hash` must contain `deviceId` and `name`.

        Note: If the info `Hash` contains `schema`, a schema is returned
        as well.
        """
        deviceId = info["deviceId"]
        name = info["name"]

        item = self.db.get_configuration(deviceId, name)
        if not item:
            reason = (f"Failure: No configuration for device {deviceId} and "
                      f"name {name} found!")
            raise KaraboError(reason)

        item = dictToHash(item)
        config64 = item["config"]
        schema64 = item["schema"]
        item["config"] = hashFromBase64Bin(config64)
        item["schema"] = schemaFromBase64Bin(schema64)

        # A configuration is always connected to a schema, but typically
        # the majority is not interested in that. Hence, we remove it
        # under conditions here and not in the config db logic.
        if not info.has("schema") or info["schema"] is False:
            item.pop("schema")

        return Hash("success", True, "item", item)

    @slot
    async def slotGetLastConfiguration(self, info):
        """Slot to get a the last configuration

        The info `Hash` must contain `deviceId` and can obtain `priority`.

        Note: If the info `Hash` contains `schema`, a schema is returned
        as well.
        """
        deviceId = info["deviceId"]
        priority = info.get("priority", 3)

        item = self.db.get_last_configuration(deviceId, priority=priority)
        if not item:
            reason = (f"No configuration for device {deviceId} and "
                      f"priority {priority} found!")
            raise KaraboError(reason)

        item = dictToHash(item)
        config64 = item["config"]
        schema64 = item["schema"]
        item["config"] = hashFromBase64Bin(config64)
        item["schema"] = schemaFromBase64Bin(schema64)

        # A configuration is always connected to a schema, but typically
        # the majority is not interested in that. Hence, we remove it
        # under conditions here and not in the config db logic.
        if not info.has("schema") or info["schema"] is False:
            item.pop("schema")

        return Hash("success", True, "item", item)

    @slot
    async def slotCheckConfigurationFromName(self, info):
        """Slot to check configuration(s) from name
           - name: the non-empty (and unique for the device) name to be
                   associated with the configuration(s)

           - deviceIds: a vector of strings with deviceIds
        """
        config_name = info["name"]  # Note: Must be there!
        deviceIds = info["deviceIds"]
        taken = self.db.is_config_name_taken(config_name, deviceIds)

        return Hash("success", True, "taken", taken)

    @slot
    async def slotListDevices(self, info):
        """List deviceIds that are stored in the database for a priority

        The info hash must contain the priority!
        """
        priority = info.get("priority", 3)
        devices = self.db.list_devices(priority=int(priority))
        return Hash("success", True, "item", devices)

    @slot
    async def slotSaveConfigurationFromName(self, info):
        """Slot to save configuration(s) from name

          The info Hash should contain:

           - client: client information
           - name: the non-empty (and unique for the device) name to be
                   associated with the configuration(s)

           - priority: priority of the configuration(s) as integer ranging
                       from 1-3 with 3 being the highest priority

           - description: an optional description for the named configs

           - user: the user name of the currently logged in user in the client

           - deviceIds: a vector of strings with deviceIds

           - overwritable: Will the configuration be overwritable?
                           Ignored if the named configuration already exists

        """
        config_name = info["name"]  # Note: Must be there!
        deviceIds = info["deviceIds"]
        if not self.name_pattern.match(config_name):
            raise KaraboError(f"The config name {config_name} does not "
                              f"comply with the allowed settings. Don't use "
                              f"special character or spaces and the maximum "
                              f"size is 30 characters.")

        if len(deviceIds) > int(self.confBulkLimit):
            raise KaraboError(
                f"The number of configurations {len(deviceIds)}"
                f" exceeds the allowed limit {self.confBulkLimit}")

        is_taken = self.db.is_config_name_taken(config_name, deviceIds)
        if is_taken:
            raise KaraboError(f"The config name {config_name} is already "
                              "used by a read-only configuration saved "
                              "for at least one of the devices in the list: "
                              f"{deviceIds}. Or by one writable configuration "
                              "that doesn't match the same list of devices.")

        # Optional information that can be taken as defaults!
        user = info.get("user", ".")
        description = info.get("description", "")
        priority = info.get("priority", 1)
        timestamp = info.get("timestamp", Timestamp().toLocal())
        overwritable = info.get("overwritable", False)

        try:
            async def poll_(device_id):
                schema, _ = await self.call(device_id, "slotGetSchema", False)
                config, _ = await self.call(device_id, "slotGetConfiguration")
                config_dict = {}
                config64 = hashToBase64Bin(config)
                schema64 = schemaToBase64Bin(schema)
                config_dict.update({"deviceId": device_id})
                config_dict.update({"config": config64})
                config_dict.update({"schema": schema64})
                return config_dict

            futures = [poll_(device_id) for device_id in deviceIds]
            timeout = len(deviceIds) * DEVICE_TIMEOUT
            items = await wait_for(gather(*futures), timeout=timeout)
        except (CancelledError, TimeoutError):
            raise

        # Let it throw here if needed!
        self.db.save_configuration(
            config_name, items, description=description, user=user,
            priority=priority, overwritable=overwritable, timestamp=timestamp)

        return Hash("success", True)

    @slot
    async def slotInstantiateDevice(self, info):
        """Slot to instantiate a device via the configuration manager

        The required info `Hash` must have at least the params:

        - deviceId: The mandatory parameter
        - name: Optional parameter. If no `name` is provided, the latest
                configuration is retrieved with priority 3 (INIT).
        - classId: Optional parameter for validation
        - serverId: Optional parameter
        """
        deviceId = info["deviceId"]
        name = info.get("name", None)
        # Note: classId and serverId can be optional
        # classId is taken for validation!
        classId = info.get("classId", None)
        serverId = info.get("serverId", None)

        # If no `name` is provided, we try to get the latest config
        if name is None:
            item = self.db.get_last_configuration(deviceId, priority=3)
            if not item:
                reason = (f"No configuration for device {deviceId} and "
                          f"priority 3 (INIT) found!")
                raise KaraboError(reason)
        else:
            item = self.db.get_configuration(deviceId, name)
            if not item:
                reason = (f"No configuration for device {deviceId} and name "
                          f"{name} found!")
                raise KaraboError(reason)
        config = hashFromBase64Bin(item["config"])
        # If the classId was provided we can validate!
        if classId is not None and classId != config["classId"]:
            raise KaraboError(f"The configuration for {deviceId} was "
                              f"recorded for classId {classId}. The "
                              f"input classId {classId} does not match!")
        # If we did not provide the classId, we take it from the configuration!
        if classId is None:
            classId = config["classId"]
        # If we did not provide a serverId, we take it from the config
        if serverId is None:
            serverId = config["serverId"]
        # Get the class schema for this device!
        schema = self._class_schemas[serverId].get(classId, None)
        if schema is None:
            # No schema there, check if we are already looking for it!
            future = self._schema_futures.get((serverId, classId), None)
            if future is not None:
                # Let it throw here in case of failure
                await future
                schema = future.result()
            else:
                future = Future()
                self._schema_futures[(serverId, classId)] = future
                try:
                    schema, *_ = await wait_for(
                        self.call(serverId, "slotGetClassSchema", classId),
                        timeout=DEVICE_TIMEOUT)
                except (CancelledError, TimeoutError) as e:
                    future.set_exception(e)
                    raise KaraboError(
                        f"server {serverId} is not available to start "
                        f"device with deviceId {deviceId} ... Could not "
                        f"retrieve the schema!")
                except Exception as e:
                    # In case we experience an unknown error, notify!
                    future.set_exception(e)
                    raise KaraboError(
                        f"Exception occured when starting device with "
                        f"deviceId {deviceId} on {serverId}: {e}")
                else:
                    future.set_result(schema)
                    # Got a new schema, cache it!
                    self._class_schemas[serverId][classId] = schema
                finally:
                    self._schema_futures.pop((serverId, classId), None)

        config = sanitize_init_configuration(schema, config)
        h = Hash()
        h["deviceId"] = deviceId
        h["classId"] = classId
        h["serverId"] = serverId
        h["configuration"] = config

        success, msg = await self.call(serverId, "slotStartDevice", h)
        if not success:
            raise KaraboError(msg)

        return Hash("success", True)


def get_scene(deviceId):
    scene0 = TableElementModel(
        height=431.0, keys=[f'{deviceId}.view'],
        parent_component='DisplayComponent',
        width=601.0, x=340.0, y=230.0)
    scene1 = DisplayTextLogModel(
        height=199.0, keys=[f'{deviceId}.status'],
        parent_component='DisplayComponent',
        width=531.0, x=410.0, y=10.0)
    scene20 = DisplayCommandModel(
        height=35.0, keys=[f'{deviceId}.listConfigurations'],
        parent_component='DisplayComponent', width=321.0, x=80.0, y=311.0)
    scene21 = DisplayCommandModel(
        height=35.0, keys=[f'{deviceId}.saveConfigurations'],
        parent_component='DisplayComponent', width=321.0, x=80.0, y=346.0)
    scene2 = BoxLayoutModel(
        direction=2, height=80.0, width=321.0, x=10.0, y=320.0,
        children=[scene20, scene21])
    scene3 = LabelModel(
        font='Source Sans Pro,14,-1,5,50,0,1,0,0,0', height=31.0,
        parent_component='DisplayComponent', text='Configuration Manager',
        width=391.0, x=20.0, y=10.0)
    scene40 = DisplayLabelModel(
        font_size=10, height=31.0, keys=[f'{deviceId}.deviceId'],
        parent_component='DisplayComponent', width=311.0, x=70.0, y=50.0)
    scene41 = DisplayStateColorModel(
        height=31.0, keys=[f'{deviceId}.state'],
        parent_component='DisplayComponent',
        show_string=True, width=311.0, x=70.0, y=90.0)
    scene420 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=31.0, parent_component='DisplayComponent', text='Last Success',
        width=73.0, x=70.0, y=130.0)
    scene421 = ErrorBoolModel(
        height=31.0, keys=[f'{deviceId}.lastSuccess'],
        parent_component='DisplayComponent', width=68.0, x=143.0, y=130.0)
    scene42 = BoxLayoutModel(
        height=37.0, width=391.0, x=10.0, y=114.0,
        children=[scene420, scene421])
    scene4 = BoxLayoutModel(
        direction=2, height=111.0, width=391.0, x=10.0, y=40.0,
        children=[scene40, scene41, scene42])
    scene5 = LineModel(
        stroke='#000000', x1=30.0, x2=900.0, y1=210.0, y2=210.0)
    scene6 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0,Normal', height=31.0,
        parent_component='DisplayComponent', text='Save options', width=161.0,
        x=10.0, y=410.0)
    scene7 = LineModel(
        stroke='#000000', x1=10.0, x2=330.0, y1=450.0, y2=450.0)
    scene800 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=20.0, parent_component='DisplayComponent',
        text='configurationName', width=321.0, x=10.0, y=470.0)
    scene8010 = DisplayLabelModel(
        font_size=10, height=27.0,
        keys=[f'{deviceId}.configurationName'],
        parent_component='DisplayComponent', width=161.0, x=10.0, y=490.0)
    scene8011 = EditableRegexModel(
        height=27.0, keys=[f'{deviceId}.configurationName'],
        parent_component='EditableApplyLaterComponent', width=160.0, x=171.0,
        y=490.0)
    scene801 = BoxLayoutModel(
        height=37.0, width=321.0, x=10.0, y=490.0,
        children=[scene8010, scene8011])
    scene80 = BoxLayoutModel(
        direction=2, height=57.0, width=321.0, x=10.0, y=470.0,
        children=[scene800, scene801])
    scene810 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=20.0, parent_component='DisplayComponent', text='priority',
        width=321.0, x=10.0, y=530.0)
    scene8110 = DisplayLabelModel(
        font_size=10, height=34.0, keys=[f'{deviceId}.priority'],
        parent_component='DisplayComponent', width=161.0, x=10.0, y=550.0)
    scene8111 = IntLineEditModel(
        height=34.0, keys=[f'{deviceId}.priority'],
        parent_component='EditableApplyLaterComponent', width=160.0, x=171.0,
        y=550.0)
    scene811 = BoxLayoutModel(
        height=37.0, width=321.0, x=10.0, y=547.0,
        children=[scene8110, scene8111])
    scene81 = BoxLayoutModel(
        direction=2, height=57.0, width=321.0, x=10.0, y=527.0,
        children=[scene810, scene811])
    scene820 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=20.0, parent_component='DisplayComponent', text='description',
        width=321.0, x=10.0, y=584.0)
    scene8210 = DisplayLabelModel(
        font_size=10, height=34.0, keys=[f'{deviceId}.description'],
        parent_component='DisplayComponent', width=161.0, x=10.0, y=604.0)
    scene8211 = LineEditModel(
        height=34.0, keys=[f'{deviceId}.description'],
        klass='EditableLineEdit',
        parent_component='EditableApplyLaterComponent', width=160.0, x=171.0,
        y=604.0)
    scene821 = BoxLayoutModel(
        height=37.0, width=321.0, x=10.0, y=604.0,
        children=[scene8210, scene8211])
    scene82 = BoxLayoutModel(
        direction=2, height=57.0, width=321.0, x=10.0, y=584.0,
        children=[scene820, scene821])
    scene8 = BoxLayoutModel(
        direction=2, height=171.0, width=321.0, x=10.0, y=470.0,
        children=[scene80, scene81, scene82])
    scene90 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=20.0, parent_component='DisplayComponent', text='Device',
        width=311.0, x=10.0, y=220.0)
    scene91 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.deviceName'],
        parent_component='DisplayComponent', width=311.0, x=10.0, y=240.0)
    scene92 = LineEditModel(
        height=27.0, keys=[f'{deviceId}.deviceName'],
        klass='EditableLineEdit',
        parent_component='EditableApplyLaterComponent', width=311.0, x=10.0,
        y=270.0)
    scene9 = BoxLayoutModel(
        direction=2, height=91.0, width=321.0, x=10.0, y=220.0,
        children=[scene90, scene91, scene92])
    scene = SceneModel(
        height=673.0, width=948.0,
        children=[scene0, scene1, scene2, scene3, scene4, scene5, scene6,
                  scene7, scene8, scene9])
    return write_scene(scene)
